"""
The functionality of bio_embeddings is split into 5 different modules

.. autosummary::
   bio_embeddings.embed
   bio_embeddings.extract
   bio_embeddings.project
   bio_embeddings.utilities
   bio_embeddings.visualize
"""
