from bio_embeddings.extract.annotations.Disorder import Disorder
from bio_embeddings.extract.annotations.Location import Location
from bio_embeddings.extract.annotations.Membrane import Membrane
from bio_embeddings.extract.annotations.SecondaryStructure import SecondaryStructure

__all__ = ["Disorder", "Location", "Membrane", "SecondaryStructure"]
